##
##

import logging
import botocore.exceptions
from typing import Union, List
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import AWSTagStruct, AWSTag

logger = logging.getLogger('couchformation.aws.driver.nsg')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class SecurityGroup(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, vpc_id: str, filter_keys_exist: Union[List[str], None] = None) -> List[dict]:
        sg_list = []
        sgs = []
        extra_args = {}
        vpc_filter = {
            'Name': 'vpc-id',
            'Values': [
                vpc_id,
            ]
        }

        try:
            while True:
                result = self.ec2_client.describe_security_groups(**extra_args, Filters=[vpc_filter])
                sgs.extend(result['SecurityGroups'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting security groups: {err}")

        for sg_entry in sgs:
            sg_block = {'name': sg_entry['GroupName'],
                        'description': sg_entry['Description'],
                        'id': sg_entry['GroupId'],
                        'vpc': sg_entry['VpcId']}
            if filter_keys_exist:
                if not all(key in sg_block for key in filter_keys_exist):
                    continue
            sg_list.append(sg_block)

        if len(sg_list) == 0:
            raise EmptyResultSet(f"no security groups found")
        else:
            return sg_list

    def create(self, name: str, description: str, vpc_id: str, tags: Union[dict, None] = None) -> str:
        sg_tag = AWSTagStruct.build("security-group")
        sg_tag.add(AWSTag("Name", name))
        if tags:
            for k, v in tags.items():
                sg_tag.add(AWSTag(k, str(v)))
        try:
            result = self.ec2_client.create_security_group(GroupName=name, Description=description, VpcId=vpc_id, TagSpecifications=[sg_tag.as_dict])
        except Exception as err:
            raise AWSDriverError(f"error creating security group: {err}")

        return result['GroupId']

    def add_egress(self, sg_id: str, protocol: str, from_port: int, to_port: int, cidr: str):
        try:
            result = self.ec2_client.authorize_security_group_egress(
                GroupId=sg_id,
                IpPermissions=[{
                    'IpProtocol': protocol,
                    'FromPort': from_port,
                    'ToPort': to_port,
                    'IpRanges': [{
                        'CidrIp': cidr
                    }],
                }]
            )
        except Exception as err:
            raise AWSDriverError(f"error adding egress to security group: {err}")

        return result['Return']

    def add_ingress(self, sg_id: str, protocol: str, from_port: int, to_port: int, cidr: str):
        try:
            result = self.ec2_client.authorize_security_group_ingress(
                GroupId=sg_id,
                IpPermissions=[{
                    'IpProtocol': protocol,
                    'FromPort': from_port,
                    'ToPort': to_port,
                    'IpRanges': [{
                        'CidrIp': cidr
                    }],
                }]
            )
        except Exception as err:
            raise AWSDriverError(f"error adding egress to security group: {err}")

        return result['Return']

    def delete(self, sg_id: str) -> None:
        sg = self.details(sg_id)
        if not sg:
            return
        try:
            self.ec2_client.delete_security_group(GroupId=sg_id)
        except Exception as err:
            raise AWSDriverError(f"error deleting security group: {err}")

    def get(self, name: str):
        get_filter = {
            'Name': "tag:Name",
            'Values': [
                name,
            ]
        }
        try:
            result = self.ec2_client.describe_security_groups(Filters=[get_filter])
            return result.get('SecurityGroups', [])[0]['GroupId']
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting security group details: {err}")

    def search(self, name: str):
        get_filter = {
            'Name': "tag:Name",
            'Values': [
                name,
            ]
        }
        try:
            result = self.ec2_client.describe_security_groups(Filters=[get_filter])
            return [dict({tag['Key']: tag['Value'] for tag in e.get('Tags', [])}, id=e.get('GroupId')) for e in result.get('SecurityGroups', [])]
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting security group details: {err}")

    def details(self, sg_id: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_security_groups(GroupIds=[sg_id])
            sg_entry = result['SecurityGroups'][0]
            sg_block = {'name': sg_entry['GroupName'],
                        'description': sg_entry['Description'],
                        'id': sg_entry['GroupId'],
                        'vpc': sg_entry['VpcId']}
            return sg_block
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting security group details: {err}")



##
##

from typing import List, Optional, Union
import attr
import yaml
import couchformation.constants as C


@attr.s
class VolumeSpec:
    v_type: str = attr.ib()
    directory: str = attr.ib()
    size: Optional[int] = attr.ib()
    run: Optional[str] = attr.ib()
    command: Optional[str] = attr.ib()


@attr.s
class ContainerSpec:
    name: str = attr.ib()
    builds: List[str] = attr.ib()
    volume: Optional[VolumeSpec] = attr.ib(default=None)
    ports: Optional[str] = attr.ib(default="80,443")

    def add_volume(self, v_type: str, directory: str, size: Union[int, None] = None, run: Union[str, None] = None, command: Union[str, None] = None):
        self.volume = VolumeSpec(v_type, directory, size, run, command)

    def add_ports(self, ports: str):
        self.ports = ports


@attr.s
class ContainerSet:
    profiles: List[ContainerSpec] = attr.ib(default=[])

    def add(self, p: ContainerSpec):
        self.profiles.append(p)

    def get(self, image):
        return next((p for p in self.profiles if p.name == image), None)

    def get_by_build(self, build):
        return next((p for p in self.profiles if build in p.builds), None)


class ContainerProfile(object):

    def __init__(self):
        self.cfg_file = C.CONTAINER_PROFILES
        self.config = ContainerSet()
        self.load_config()

    def get(self, image) -> Union[ContainerSpec, None]:
        profile = self.config.get(image)
        if not profile:
            return ContainerSpec(image, [])
        return profile

    def by_build(self, build: str) -> Union[ContainerSpec, None]:
        profile = self.config.get_by_build(build)
        if not profile:
            return None
        return profile

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for image, settings in yaml.safe_load(f).items():
                    profile = ContainerSpec(image, settings.get('builds').split(','))
                    if settings.get('ports'):
                        profile.add_ports(settings.get('ports'))
                    if settings.get('volume'):
                        vol_spec = settings.get('volume')
                        profile.add_volume(vol_spec.get('type'), vol_spec.get('directory'), vol_spec.get('size'), vol_spec.get('run'), vol_spec.get('command'))
                    self.config.add(profile)
            except yaml.YAMLError as err:
                RuntimeError(f"Can not open target config file {self.cfg_file}: {err}")



##
##

import logging
import docker
from docker import APIClient
from couchformation.exception import FatalError, NonFatalError

logger = logging.getLogger('couchformation.docker.driver.base')
logger.addHandler(logging.NullHandler())
logging.getLogger("docker").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class DockerDriverError(FatalError):
    pass


class EmptyResultSet(NonFatalError):
    pass


class CloudBase(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.project = parameters.get('project')
        self.name = parameters.get('name')

        self.client = docker.from_env()
        self.docker_api = APIClient(base_url='unix://var/run/docker.sock')
        stats = self.docker_api.version()
        eng = next((c for c in stats.get('Components', [{}]) if c.get('Name') == 'Engine'), None)

        self.arch = eng.get('Details', {}).get('Arch')

    @staticmethod
    def test_session():
        try:
            docker_api = APIClient(base_url='unix://var/run/docker.sock')
            docker_api.version()
            client = docker.from_env()
            client.info()
        except Exception as err:
            raise DockerDriverError(f"not authorized: {err}")

    @property
    def architecture(self):
        return self.arch



##
##

import logging
import re
from typing import List, Union
from couchformation.azure.driver.base import CloudBase, AzureDriverError, EmptyResultSet
from couchformation.azure.driver.constants import AzureImagePublishers
import couchformation.constants as C

logger = logging.getLogger('couchformation.azure.driver.image')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class Image(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, filter_keys_exist: Union[List[str], None] = None, resource_group: Union[str, None] = None) -> List[dict]:
        image_list = []
        if not resource_group:
            resource_group = self.rg_switch()

        images = self.compute_client.images.list_by_resource_group(resource_group)

        for image in list(images):
            image_block = {'name': image.name,
                           'id': image.id,
                           'resource_group': resource_group,
                           'location': image.location}
            image_block.update(self.process_tags(image.tags))
            if filter_keys_exist:
                if not all(key in image_block for key in filter_keys_exist):
                    continue
            image_list.append(image_block)

        if len(image_list) == 0:
            raise EmptyResultSet(f"no images found")

        return image_list

    def public(self, location: str, publisher: str, architecture: str = 'x86_64'):
        offer_list = []
        image_list = []

        offers = self.compute_client.virtual_machine_images.list_offers(location, publisher)
        for group in list(offers):
            offer_block = {'name': group.name,
                           'skus': [],
                           'count': 0}
            offer_list.append(offer_block)

        for n, offer in enumerate(offer_list):
            offer_name = offer['name']
            skus = self.compute_client.virtual_machine_images.list_skus(location, publisher, offer_name)
            for sku in list(skus):
                sku_name = sku.name
                _image_is_arm = re.search('arm64', offer_name) or re.search('arm64', sku_name)
                if architecture == 'arm64' and not _image_is_arm:
                    continue
                elif _image_is_arm:
                    continue
                versions = self.compute_client.virtual_machine_images.list(location, publisher, offer_name, sku_name)
                for version in versions:
                    image_block = {
                        "publisher": publisher,
                        "offer": offer_name,
                        "sku": sku_name,
                        "version": version.name,
                        "location": location
                    }
                    image_list.append(image_block)

        if len(image_list) == 0:
            raise EmptyResultSet(f"no images found")

        return image_list

    def details(self, name: str, resource_group: Union[str, None] = None) -> dict:
        if not resource_group:
            resource_group = self.rg_switch()
        request = self.compute_client.images.get(resource_group, name)
        image = request.result()
        image_block = {'name': image.name,
                       'id': image.id,
                       'resource_group': resource_group,
                       'location': image.location}
        image_block.update(self.process_tags(image.tags))
        return image_block

    def delete(self, name: str, resource_group: Union[str, None] = None) -> None:
        if not resource_group:
            resource_group = self.rg_switch()
        try:
            request = self.compute_client.images.begin_delete(resource_group, name)
            request.result()
        except Exception as err:
            raise AzureDriverError(f"can not delete image: {err}")

    def list_standard(self, os_id: str = None, os_version: str = None, architecture: str = 'x86_64'):
        result_list = []
        for image_type in AzureImagePublishers.publishers:
            if os_id and image_type['os_id'] != os_id:
                continue
            image_list = self.public(location=self.region, publisher=image_type['name'], architecture=architecture)
            for version in C.OS_VERSION_LIST[image_type['os_id']]:
                if os_version and version != os_version:
                    continue
                filtered_images = []
                for image in image_list:
                    offer_match = re.search(image_type['offer_match'], image['offer'])
                    sku_match = re.search(image_type['sku_match'], image['sku'])
                    if offer_match:
                        if len(offer_match.groups()) > 0:
                            m = offer_match.group(1)
                            if m == version or m == version.replace('.', '_'):
                                filtered_images.append(image)
                        elif sku_match and len(sku_match.groups()) > 0:
                            m = sku_match.group(1)
                            if m == version or m == version.replace('.', '_'):
                                filtered_images.append(image)
                if len(filtered_images) > 0:
                    filtered_images.sort(key=lambda i: i['version'])
                    result_image = filtered_images[-1]
                    result_image.update(dict(
                        os_id=image_type['os_id'],
                        os_version=version,
                        os_user=image_type['user'],
                    ))
                    result_list.append(result_image)
                if len(result_list) > 0:
                    logger.debug(f"Selected image -> {result_list[-1]}")
                return result_list[-1]

    @staticmethod
    def image_user(os_id: str):
        return next((image_type['user'] for image_type in AzureImagePublishers.publishers if image_type['os_id'] == os_id), None)



##
##

import logging
from couchformation.exception import FatalError
from couchformation.capella.driver.base import CloudBase
from couchformation.config import get_state_file, get_state_dir, State
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, Synchronize
from couchformation.util import PasswordUtility
from couchformation.deployment import MetadataManager
from couchformation.project import Project
from libcapella.columnar import CapellaColumnar
from libcapella.columnar_allowed_cidr import ColumnarAllowedCIDR
from libcapella.database import CapellaDatabase
from libcapella.database_allowed_cidr import CapellaAllowedCIDR
from libcapella.database_credentials import CapellaDatabaseCredentials
from libcapella.app_service import CapellaAppService
from libcapella.network_peers import CapellaNetworkPeers
from libcapella.logic.columnar import CapellaColumnarBuilder
from libcapella.logic.allowed_cidr import AllowedCIDRBuilder
from libcapella.logic.database import CapellaDatabaseBuilder
from libcapella.logic.credentials import DatabaseCredentialsBuilder
from libcapella.logic.app_service import CapellaAppServiceBuilder
from libcapella.logic.network_peers import NetworkPeerBuilder

logger = logging.getLogger('couchformation.capella.node')
logger.addHandler(logging.NullHandler())
logging.getLogger("restfull").setLevel(logging.ERROR)


class CapellaNodeError(FatalError):
    pass


class CapellaDeployment(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.deploy_type = parameters.get('type') if parameters.get('type') else "database"
        self.cluster_id = parameters.get('instance_id')
        self.cluster_name = parameters.get('instance_name')
        self.project_name = parameters.get('project')
        self.build = parameters.get('build')
        self.region = parameters.get('region') if parameters.get('region') else "us-east-1"
        self.cloud = parameters.get('cloud')
        self.machine_type = parameters.get('machine_type')
        self.quantity = parameters.get('quantity')
        self.provider = parameters.get('provider') if parameters.get('provider') else "aws"
        self.username = parameters.get('username') if parameters.get('username') else "Administrator"
        self.password = parameters.get('password')
        self.account_email = parameters.get('account_email')
        self.profile = parameters.get('profile') if parameters.get('profile') else 'default'
        self.sw_version = self.parameters.get('sw_version') if self.parameters.get('sw_version') else "latest"
        self.cidr = parameters.get('cidr')
        self.allow = parameters.get('allow') if parameters.get('allow') else "0.0.0.0/0"
        self.peer_project = parameters.get('peer_project')
        self.peer_region = parameters.get('peer_region')
        self.node_name = f"{self.name}-node-01"

        self.state_file = get_state_file(self.project_name, self.name)
        self.state_dir = get_state_dir(self.project_name, self.name)

        with Synchronize():
            try:
                FileManager().make_dir(self.state_dir)
            except Exception as err:
                raise CapellaNodeError(f"can not create state dir: {err}")

        document = self.node_name
        self.state = KeyValueStore(self.state_file, document)

        self.base = CloudBase(self.parameters)
        self.base.test_session()

        self.project_id = self.base.project_id
        self.org_id = self.base.organization_id
        self.project = self.base.project

    def compose(self):
        if self.deploy_type == "mobile":
            self.compose_app_svc()
        else:
            self.compose_database()

    def compose_app_svc(self):
        number = self.parameters.get('number') if self.parameters.get('number') else 1
        document = f"{self.name}-node-group-{number:02d}"
        group = KeyValueStore(self.state_file, document)

        group['machine_type'] = self.parameters.get('machine_type')
        group['quantity'] = self.parameters.get('quantity') if self.parameters.get('quantity') else 2

    def compose_database(self):
        number = self.parameters.get('number') if self.parameters.get('number') else 1
        document = f"{self.name}-node-group-{number:02d}"
        group = KeyValueStore(self.state_file, document)

        group['cloud'] = self.parameters.get('provider')
        group['machine_type'] = self.parameters.get('machine_type')
        group['volume_size'] = self.parameters.get('volume_size') if self.parameters.get('volume_size') else "256"
        group['quantity'] = self.parameters.get('quantity') if self.parameters.get('quantity') else 3
        group['services'] = self.parameters.get('services') if self.parameters.get('services') else "data,index,query"

    def deploy(self):
        if self.build == "columnar":
            self.deploy_columnar()
        else:
            if self.deploy_type == "mobile":
                self.deploy_app_svc()
            else:
                self.deploy_database()
        if self.peer_project:
            self.peer_cluster()

    def deploy_app_svc(self):
        document = f"{self.name}-node-group-01"
        group = KeyValueStore(self.state_file, document)
        self.state['state'] = State.DEPLOYING.value

        self.state['project_id'] = self.project_id
        self.state['project'] = self.project_name
        self.state['type'] = self.deploy_type

        if not self.cluster_id:
            raise CapellaNodeError(f"Please connect the app service {self.name} to a Capella database")

        self.state['instance_id'] = self.cluster_id

        quantity = int(group.get('quantity'))
        machine = group.get('machine_type')

        try:
            database = CapellaDatabase(self.project, self.cluster_name)
            app_service = CapellaAppService(database)
            if not app_service.id:
                builder = CapellaAppServiceBuilder()
                builder.name(self.name)
                builder.compute(machine, quantity)
                config = builder.build()
                logger.info(f"Creating app service {self.name} with {quantity} {machine} nodes")
                app_service.create(config)
                logger.info("Waiting for app service creation to complete")
                if not app_service.wait("healthy", until=True):
                    raise CapellaNodeError("Timeout waiting for app service to deploy")
            else:
                logger.info(f"App service {self.name} already exists")
        except Exception as err:
            raise CapellaNodeError(f"Error creating app service: {err}")

        self.state['name'] = self.name
        self.state['cluster_name'] = self.cluster_name
        self.state['app_svc_id'] = app_service.id

        logger.info(f"App service ID: {app_service.id}")

        logger.info("Capella app service successfully created")

        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def deploy_database(self):
        state_db = KeyValueStore(self.state_file)
        node_groups = state_db.doc_id_startswith(f"{self.name}-node-group")
        self.state['state'] = State.DEPLOYING.value

        if len(node_groups) == 0:
            raise CapellaNodeError("no node groups present")

        self.state['project'] = self.project_name
        self.state['type'] = self.deploy_type

        try:
            database = CapellaDatabase(self.project, self.name)
            if not database.id:
                logger.info(f"Creating cluster {self.name}")
                builder = CapellaDatabaseBuilder(self.provider)
                builder = builder.name(self.name)
                builder = builder.description("Pytest created cluster")
                builder = builder.region(self.region)
                if self.cidr:
                    builder = builder.cidr(self.cidr)
                for group in node_groups:
                    group_db = KeyValueStore(self.state_file, group)
                    machine_type = group_db.get('machine_type')
                    quantity = group_db.get('quantity')
                    volume_size = group_db.get('volume_size')
                    services = group_db.get('services').split(',')
                    builder = builder.service_group(machine_type, quantity, volume_size, services)
                config = builder.build()
                database.create(config)
                logger.info("Waiting for cluster creation to complete")
                if not database.wait("deploying"):
                    raise CapellaNodeError("Timeout waiting for cluster to deploy")
            else:
                logger.info(f"Database {self.name} already exists")

            database.refresh()
            self.state['instance_id'] = database.id
            self.state['provider'] = database.this.cloudProvider.type
            self.state['region'] = database.this.cloudProvider.region
            self.state['cidr'] = database.this.cloudProvider.cidr
            self.state['name'] = database.this.name
            self.state['cloud'] = self.cloud
            self.state['connect_string'] = database.this.connectionString

            logger.info(f"Cluster ID: {database.id}")
            logger.info(f"Cloud provider: {database.this.cloudProvider.type}")
            logger.info(f"Cloud region: {database.this.cloudProvider.region}")
            logger.info(f"Cluster CIDR: {database.this.cloudProvider.cidr}")
            logger.info(f"Connect string: {database.this.connectionString}")

            allowed_cidr = CapellaAllowedCIDR(database, self.allow)
            if not allowed_cidr.id:
                logger.info(f"Configuring allowed CIDR {self.allow}")
                builder = AllowedCIDRBuilder()
                builder.cidr(self.allow)
                config = builder.build()
                allowed_cidr.create(config)
            else:
                logger.info(f"Allow list already set to {self.state.get('allow')}")

            self.state['allow'] = self.allow

            if self.password:
                password = self.password
            else:
                password = PasswordUtility().generate(16)

            database_credential = CapellaDatabaseCredentials(database, self.username)
            if not database_credential.id:
                logger.info(f"Creating database user {self.username}")
                builder = DatabaseCredentialsBuilder(self.username, password)
                builder.data_read_write()
                config = builder.build()
                database_credential.create(config)
                self.state['password'] = password
            else:
                logger.info(f"Database user {self.state.get('username')} already exists")
        except Exception as err:
            raise CapellaNodeError(f"Error creating database: {err}")

        self.state['username'] = self.username

        logger.info("Capella database successfully created")

        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def deploy_columnar(self):
        self.state['state'] = State.DEPLOYING.value
        self.state['project'] = self.project_name
        self.state['project_id'] = self.project_id

        try:
            cluster = CapellaColumnar(self.project, self.name)
            if not cluster.id:
                logger.info(f"Creating Columnar cluster {self.name}")
                builder = CapellaColumnarBuilder(self.provider)
                builder = builder.name(self.name)
                builder = builder.description("Pytest created cluster")
                builder = builder.region(self.region)
                builder = builder.compute(self.machine_type, self.quantity)
                config = builder.build()
                cluster.create(config)
                logger.info("Waiting for cluster creation to complete")
                if not cluster.wait("deploying"):
                    raise CapellaNodeError("Timeout waiting for cluster to deploy")
            else:
                logger.info(f"Columnar cluster {self.name} already exists")

            logger.info(f"Cluster ID: {cluster.id}")

            cluster.refresh()
            self.state['instance_id'] = cluster.id
            self.state['provider'] = cluster.this.cloudProvider
            self.state['region'] = cluster.this.region
            self.state['name'] = cluster.this.name
            self.state['cloud'] = self.cloud

            allowed_cidr = ColumnarAllowedCIDR(cluster, self.allow)
            if not allowed_cidr.id:
                logger.info(f"Configuring allowed CIDR {self.allow}")
                builder = AllowedCIDRBuilder()
                builder.cidr(self.allow)
                config = builder.build()
                allowed_cidr.create(config)
            else:
                logger.info(f"Allow list already set to {self.allow}")
        except Exception as err:
            raise CapellaNodeError(f"Error creating cluster: {err}")

        self.state['allow'] = self.allow

        logger.info("Columnar cluster successfully created")

        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def peer_cluster(self):
        peer_project = self.peer_project
        if not MetadataManager(peer_project).exists:
            logger.warning(f"Can not peer with project {peer_project}: project does not exist")
            return

        try:
            peer_region = self.peer_region if self.peer_region else self.region
            database = CapellaDatabase(self.project, self.name)
            state_data = MetadataManager(peer_project).get_network_state(self.provider, peer_region)
            parameters = MetadataManager(peer_project).get_network_params(self.provider, peer_region)

            if state_data.get('state') != State.DEPLOYED.value:
                logger.warning(f"Project network is not deployed. Peering aborted.")
                return

            network_peer = CapellaNetworkPeers(database)
            if not network_peer.id:
                if self.provider == "aws":
                    account_id = state_data.get('account_id')
                    vpc_id = state_data.get('vpc_id')
                    vpc_cidr = state_data.get('vpc_cidr')
                    builder = NetworkPeerBuilder()
                    builder.account_id(account_id)
                    builder.vpc_id(vpc_id)
                    builder.region(self.region)
                    builder.cidr(vpc_cidr)
                    config = builder.build()
                    network_peer.create(config)
                elif self.provider == "gcp":
                    gcp_project_name = state_data.get('project_name')
                    network_name = state_data.get('network')
                    network_cidr = state_data.get('network_cidr')
                    service_account = state_data.get('default_service_account')
                    builder = NetworkPeerBuilder()
                    builder.provider_type("gcp")
                    builder.network_name(network_name)
                    builder.cidr(network_cidr)
                    builder.project_id(gcp_project_name)
                    builder.service_account(service_account)
                    config = builder.build()
                    network_peer.create(config)
                elif self.provider == "azure":
                    raise CapellaNodeError(f"Can not peer with project {self.provider}: Azure is currently not supported")
                else:
                    raise CapellaNodeError(f"Can not peer with project {self.provider}: unsupported cloud provider {self.provider}")

            network_peer.refresh()

            self.state['provider_id'] = network_peer.provider_id
            parameters['provider_id'] = self.state['provider_id']
            logger.info(f"Capella peer provider ID: {self.state['provider_id']}")

            if network_peer.hosted_zone_id:
                self.state['aws_hosted_zone'] = network_peer.hosted_zone_id
                parameters['hosted_zone'] = self.state['aws_hosted_zone']
                logger.info(f"Capella AWS hosted zone ID: {self.state['aws_hosted_zone']}")

            if network_peer.peer_project:
                self.state['gcp_peer_project'] = network_peer.peer_project
                parameters['peer_gcp_project'] = self.state['gcp_peer_project']
                logger.info(f"Capella GCP project: {self.state['gcp_peer_project']}")

            if network_peer.peer_network:
                self.state['gcp_peer_network'] = network_peer.peer_network
                parameters['peer_gcp_network'] = self.state['gcp_peer_network']
                logger.info(f"Capella GCP network: {self.state['gcp_peer_network']}")

            if network_peer.managed_zone:
                self.state['gcp_managed_zone'] = network_peer.managed_zone
                parameters['managed_gcp_zone'] = self.state['gcp_managed_zone']
                logger.info(f"Capella GCP managed zone: {self.state['gcp_managed_zone']}")

        except Exception as err:
            raise CapellaNodeError(f"Error peering network: {err}")

        self.state['network_peer_id'] = network_peer.id

        logger.info(f"Network Peer ID: {network_peer.id}")

        logger.info(f"Invoking peer acceptance module for cloud {self.provider}")
        Project.peer_network(self.provider, parameters)
        database.wait("peering")
        logger.info(f"Peering complete")

    def unpeer_project(self):
        peer_project = self.peer_project
        if not MetadataManager(peer_project).exists:
            logger.warning(f"Can not unpeer with project {peer_project}: project does not exist")
            return

        try:
            peer_region = self.peer_region if self.peer_region else self.region
            database = CapellaDatabase(self.project, self.name)
            state_data = MetadataManager(peer_project).get_network_state(self.provider, peer_region)
            parameters = MetadataManager(peer_project).get_network_params(self.provider, peer_region)

            if state_data.get('state') != State.DEPLOYED.value:
                logger.warning(f"Project network is not deployed. Peering removal aborted.")
                return

            network_peer = CapellaNetworkPeers(database)
            if network_peer.id:
                network_peer.refresh()
                logger.info(f"Removing Capella peering {network_peer.name}")
                network_peer.delete()

            if self.state.get('provider_id'):
                parameters['provider_id'] = self.state['provider_id']

            if self.state.get('aws_hosted_zone'):
                parameters['hosted_zone'] = self.state['aws_hosted_zone']

            if self.state.get('gcp_peer_project'):
                parameters['peer_gcp_project'] = self.state['gcp_peer_project']

            if self.state.get('gcp_peer_network'):
                parameters['peer_gcp_network'] = self.state['gcp_peer_network']

            if self.state.get('gcp_managed_zone'):
                parameters['managed_gcp_zone'] = self.state['gcp_managed_zone']

        except Exception as err:
            raise CapellaNodeError(f"Error unpeering network: {err}")

        logger.info(f"Invoking peer removal module for cloud {self.provider}")
        Project.unpeer_network(self.provider, parameters)
        database.wait("peering")
        logger.info(f"Peer removal complete")

    def destroy(self):
        if self.peer_project:
            self.unpeer_project()
        if self.build == "columnar":
            self.destroy_columnar()
        else:
            if self.state.get('type') == "mobile":
                self.destroy_app_svc()
            else:
                self.destroy_database()

    def destroy_app_svc(self):
        try:
            self.state['state'] = State.DESTROYING.value
            app_svc_name = self.state['name']
            cluster_name = self.state['cluster_name']
            database = CapellaDatabase(self.project, cluster_name)
            app_service = CapellaAppService(database)

            if app_service.id:
                logger.info(f"Destroying app service {app_svc_name}")
                app_service.delete()
                if not app_service.wait("destroying"):
                    raise CapellaNodeError("Timeout waiting for app service deletion to complete")
            else:
                logger.info(f"Cluster {cluster_name} does not have associated app services")

            self.state.clear()
            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise CapellaNodeError(f"Error destroying app service: {err}")

    def destroy_database(self):
        try:
            self.state['state'] = State.DESTROYING.value
            cluster_name = self.state['name'] = self.name
            database = CapellaDatabase(self.project, cluster_name)

            if database.id:
                logger.info(f"Destroying cluster {cluster_name}")
                database.delete()
                if not database.wait("destroying"):
                    raise CapellaNodeError("Timeout waiting for cluster deletion to complete")
            else:
                logger.info(f"Database {cluster_name} does not exist")

            self.state.clear()
            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise CapellaNodeError(f"Error destroying database: {err}")

    def destroy_columnar(self):
        try:
            self.state['state'] = State.DESTROYING.value
            cluster = CapellaColumnar(self.project, self.name)

            if cluster.id:
                logger.info(f"Destroying Columnar cluster {self.name}")
                cluster.delete()
                logger.info("Waiting for cluster removal to complete")
                if not cluster.wait("destroying"):
                    raise CapellaNodeError("Timeout waiting for cluster deletion to complete")
            else:
                logger.info(f"Columnar cluster {self.name} does not exist")

            logger.info("Columnar cluster successfully removed")

            self.state.clear()
            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise CapellaNodeError(f"Error destroying cluster: {err}")

    def info(self):
        return self.state.as_dict



##
##

import logging
import warnings
import argparse
import json
from overrides import override
import couchformation
from couchformation.cli.cli import CLI
from couchformation.project import Project
from couchformation.resources.config_manager import ConfigurationManager
from couchformation.support.debug import CreateDebugPackage

warnings.filterwarnings("ignore")
logger = logging.getLogger()


class CloudMgrCLI(CLI):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @override()
    def local_args(self):
        self.parser.add_argument('-V', action='store_true', dest='show_version', help="Display version information")

        opt_parser = argparse.ArgumentParser(parents=[self.parser], add_help=False)
        opt_parser.add_argument('-b', '--build', action='store', help="Build Type", default="cbs")
        opt_parser.add_argument('-c', '--cloud', action='store', help="Infrastructure", default="aws")
        opt_parser.add_argument('-p', '--project', action='store', help="Project Name")
        opt_parser.add_argument('-n', '--name', action='store', help="Service Name")
        opt_parser.add_argument('-x', '--connect', action='store', help="Connection Name", default=None)
        opt_parser.add_argument('-g', '--group', action='store', help="Group Number", default=1, type=int)
        opt_parser.add_argument('-P', '--provisioner', action='store', help="Provisioner Name", default="remote")
        opt_parser.add_argument('-R', '--raw', action='store_true', help="Skip provision phase")
        opt_parser.add_argument('-t', '--to', action='store', help="Copy target")
        opt_parser.add_argument('-E', '--extended', action='store_true', help="Extended output")
        opt_parser.add_argument('--json', action='store_true', help="List output in JSON")

        command_subparser = self.parser.add_subparsers(dest='command')
        command_subparser.add_parser('create', help="Create New Service", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('add', help="Add Resource Group", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('copy', help="Copy Project", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('deploy', help="Deploy Project", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('destroy', help="Destroy Services", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('remove', help="Remove Services", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('clean', help="Clean Project", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('list', help="List Projects", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('show', help="Project Information", parents=[opt_parser], add_help=False, aliases=['info'])
        command_subparser.add_parser('parameters', help="Show Project Parameters", parents=[opt_parser], add_help=False, aliases=['param', 'parm'])
        command_subparser.add_parser('dump', help="Create Debug Bundle", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('cli', help="Get Project Create CLI", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('update', help="Edit Service Settings", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('peer', help="Network Peering", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('login', help="Cloud Login", parents=[opt_parser], add_help=False)
        command_subparser.add_parser('help', help="Show Supported Options", parents=[opt_parser], add_help=False)

        config_cmd = command_subparser.add_parser('config', help="Configuration Manager", parents=[opt_parser], add_help=False)
        config_parser = config_cmd.add_subparsers(dest='config_command')
        config_parser.add_parser('get', help="Get Config Elements", parents=[opt_parser], add_help=False)
        config_parser.add_parser('set', help="Get Config Elements", parents=[opt_parser], add_help=False)
        config_parser.add_parser('unset', help="Get Config Elements", parents=[opt_parser], add_help=False)

    def run(self):
        if not hasattr(self.options, 'json'):
            logger.info(f"Couch Formation v{couchformation.__version__}")

        if self.options.show_version:
            return

        if self.options.command == "config":
            self.config_mgr(self.options.config_command)
            return

        if self.options.command == "dump":
            CreateDebugPackage().create_snapshot()
            return

        if self.options.command == "list" and not self.options.project:
            Project(self.options, self.remainder).list_projects()
            return

        if self.options.command == "login" and self.options.cloud:
            Project(self.options, self.remainder).login(self.options.cloud)
            return

        if self.options.command == "help":
            logger.info("General parameters:\n")
            self.parser.print_help()
            print("")
            Project(self.options, self.remainder).show_help()
            return

        if not self.options.command or not self.options.project:
            logger.error("Missing required arguments")
            self.parser.print_help()
            return

        project = Project(self.options, self.remainder)

        if self.options.command == "create":
            if self.options.name is None:
                logger.error("Missing required parameter: name")
                return
            project.create()
        elif self.options.command == "add":
            if self.options.name is None:
                logger.error("Missing required parameter: name")
                return
            project.add()
        elif self.options.command == "copy":
            if self.options.to is None:
                logger.error("Missing required parameter: to")
            project.copy()
        elif self.options.command == "deploy":
            project.deploy(self.options.name, self.options.raw)
        elif self.options.command == "destroy":
            project.destroy(self.options.name)
        elif self.options.command == "remove":
            project.remove()
        elif self.options.command == "clean":
            project.clean()
        elif self.options.command == "show" or self.options.command == "list" or self.options.command == "info":
            results = project.list(api=self.options.json)
            if self.options.json:
                print(json.dumps(results, indent=2))
        elif self.options.command == "param" or self.options.command == "parameters" or self.options.command == "parm":
            project.project_show()
        elif self.options.command == "cli":
            project.project_cli()
        elif self.options.command == "update":
            if self.options.name is None:
                logger.error("Missing required parameter: name")
                return
            project.service_edit()
        elif self.options.command == "peer":
            project.accept_peering(self.options.name)
        elif self.options.command == "login":
            project.login()

        loggers = [logging.getLogger()] + list(logging.Logger.manager.loggerDict.values())
        for log in loggers:
            handlers = getattr(log, 'handlers', [])
            for handler in handlers:
                handler.flush()

    def config_mgr(self, command: str):
        cm = ConfigurationManager()
        if command == "get":
            if len(self.remainder) == 0:
                contents = cm.list()
                for key in contents:
                    print(f"{key} = {contents[key]}")
                return
            elif len(self.remainder) == 1:
                value = cm.get(self.remainder[0])
                if value is not None:
                    print(f"{self.remainder[0]} = {value}")
            else:
                logger.error(f"Usage: get [key_name]")
        elif command == "set":
            if len(self.remainder) == 2:
                cm.set(self.remainder[0], self.remainder[1])
            else:
                logger.error(f"Usage: set key_name value")
        elif command == "unset":
            if len(self.remainder) == 1:
                cm.delete(self.remainder[0])
            else:
                logger.error(f"Usage: unset key_name")
        else:
            logger.error(f"Unknown config command: {command}")


def main(args=None):
    cli = CloudMgrCLI(args)
    cli.run()
